//////////////////////////////////////////////////////////////////////
//
// File:     DMShell.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     9/1/2003
// Requires: comshell package
//
// Description:
//
//   A command line interface to the data mining (dm) package.
//
//////////////////////////////////////////////////////////////////////

// Package definition
package dm;

// Packages to import
import comshell.*;
import dm.dmql.*;
import dm.datacube.*;
import java.io.*;
import java.lang.*;
import java.util.*;

/**
 * Text shell interface for a data mining system
 *
 * @version   1.0
 * @author    Scott Sanner
 * @language  Java (JDK 1.3)
 **/
public class DMShell
{
    /* Static constants */
    public final static int    MAX_INPUT          = 4096;
    public final static String DEFAULT_PREFS_FILE = "prefs.txt";
    public final static int    INDENT             = 3;

    /* Encapsulated Objects */
    public CommandInterface _ci;
    public InputStream      _is;
    public PrintStream      _os;
    public Domain           _domain;
    public long             _lStartTime;

    /* Class-defined commands */
    public int QUIT;
    public int COMMENT;
    public int NEW_DOMAIN;
    public int DMQL_QUERY;
    public int TIMER;
    
    /** Initializes the shell interface and invokes it
     **/
    public static void main(String args[]) 
    {
	try {

	    InputStream in;
	    if (args.length >= 1)
		in = new FileInputStream(args[0]);
	    else
		in  = System.in;
		
	    DMShell shell = new DMShell(in, System.out);
	    shell.run();

	} catch (FileNotFoundException e) 
	    {
		System.out.println("File IO Error while reading " + args[0] + ", exiting...");
		System.exit(1);
	    }
	
    } 
    
    /** Empty constructor - uses System input/output stream and default
     *  preferences file.
     **/
    public DMShell()
    {
	this(System.in, System.out, DEFAULT_PREFS_FILE);
    }

    /** Constructor - uses default preferences file.
     *  @param is InputStream to read input from
     *  @param os OutputStream to write output to
     **/
    public DMShell(InputStream is, PrintStream os)
    {
	this(is, os, DEFAULT_PREFS_FILE);
    }

    /** Constructor
     *  @param is InputStream to read input from
     *  @param os OutputStream to write output to
     *  @param prefs_file File to load default environmental variable
     *                    bindings/preferences from
     **/
    public DMShell(InputStream is, PrintStream os, String prefs_file)
    {
	_is     = is;
	_os     = os;
	_ci    = new CommandInterface(_is, _os);

	// Initialize a set of commands
	TIMER         = _ci.command.addCommand("timer",
					       " {start|stop}                  - start or stop timer");
	QUIT          = _ci.command.addCommand("quit", 
					       "                                - quit application");
	COMMENT       = _ci.command.addCommand("//", 
					       "                                  - comment\n");
	NEW_DOMAIN    = _ci.command.addCommand("new-domain", 
					       " <name>                   - make a new index");
	DMQL_QUERY    = _ci.command.addCommand("dmql-query",
					       " <string>                 - query index and rank docs");
	
	// Initialize environmental variable bindings from preferences file
	_ci.initEnvVarsFromFile(prefs_file);

	// Initialize time
	_lStartTime = 0L;

	// Initialize the domain
	String driver = _ci.getBindings("DB-DRIVER");
	String dbname = _ci.getBindings("DB-NAME");
	_domain = new Domain("Default Domain", driver, dbname);
	_os.println("\nCreated new domain default domain from db '" + dbname + "'...");
    }
    
    /** Main command line handler
     **/
    public void run()
    {
      
	while (_ci.command.type != QUIT) {

	    try {
		_ci.getCommand();
	    } catch (IOException e) {
		_os.println("IO Error: " + e);
		while (true);
	    }
		
	    /***********************************************************
	     * Command: Quit
	     ***********************************************************/
	    if (_ci.command.type == QUIT) {
		_os.println("\nExiting DMShell\n");		
	    }

	    /***********************************************************
	     * Command: Handle timer commands
	     ***********************************************************/
	    else if (_ci.command.type == TIMER) {

		if (_ci.command.numParams() != 1) {
		    _os.println("\nNeed to specify either 'start' or 'stop'.");
		} else {
		    String com = _ci.command.getParam(0);
		    if (com.equalsIgnoreCase("start")) {
			_lStartTime = System.currentTimeMillis();
		    } else if (com.equalsIgnoreCase("stop")) {
			_os.println("\nElapsed time: " + (System.currentTimeMillis() - 
							  _lStartTime) + " ms");
		    } else {
			_os.println("\nUnrecognized timer command");
		    }
		}
	    }
		
	    /***********************************************************
	     * Command: Comment '//'
	     ***********************************************************/
	    if (_ci.command.type == COMMENT) {
		; // Do nothing
	    }
		
	    /***********************************************************
	     * Command: New Domain
	     ***********************************************************/
	    else if (_ci.command.type == NEW_DOMAIN) {

		// Obtain the domain title (if provided)
		String domain_title = "Default Domain";
		if (_ci.command.numParams() >= 1) {
		    domain_title = _ci.command.getParam(0);
		}

		// Obtain the current driver and dbname environmental
		// variable bindings (use listenv from DMShell to see
		// bindings)
		String driver = _ci.getBindings("DB-DRIVER"); // Env. variable!
		String dbname = _ci.getBindings("DB-NAME");   // Env. variable!

		// Create the domain
		_domain = new Domain(domain_title, driver, dbname);

		// Print results
		_os.println("\nCreated new domain '" + domain_title + "' from db '" + dbname + "'...");
	    }

	    /***********************************************************
	     * Command: DMQL Query
	     ***********************************************************/
	    else if (_ci.command.type == DMQL_QUERY) {

		if (_ci.command.numParams() != 1) {
		    _os.println("\nMust specify a single query string.");
		} else {
		    
		    // Parse the command
		    DMQLCommand dcom = DMQL.ProcessQuery(_ci.command.getParam(0));

		    // Process the command
		    if (dcom == null) {
			
			// dcom is null so display error
			_os.println("\nDMQL query string invalid.");

		    } else { 
			
			// dcom is valid so process the query
			if (_domain.process(dcom)) {
			    _os.println("\nDMQL command successfully executed.");
			} else {
			    _os.println("\nError while processing DMQL command.");
			}
		    }
		}
	    }
	}
    }

}











