//////////////////////////////////////////////////////////////////////
//
// File:     DataCube.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     10/1/2003
//
// For reasons of simplicity, we assume that all of the dimensions 
// in the data cube are ints and that the value is an int.  Consequently
// we represent the data cube as an array of DataElement objects where
// a DataElement is a [key: int[], val: int]
//
//////////////////////////////////////////////////////////////////////

// Package definition
package dm.datacube;

// Packages to import
import dm.dmql.*;
import java.io.*;
import java.text.*;
import java.util.*;

/**
 * Implements a DataCube using an array and a DCComp object
 *
 * @version   1.0
 * @author    Scott Sanner
 * @language  Java (JDK 1.3)
 **/
public class DataCube {

    /** Constants **/
    public final static int INITIAL_SIZE = 100;

    /** Local data members **/
    public DataElement[] _data;            // The data in this cube!

    public ArrayList     _alDimensions;    // List of dimensions in the order that are
                                           // used in the DataElement key
    public Dimension     _dimVal;          // Dimension for value

    public DCComp        _comp;            // The comparator used for this DataCube
    public boolean       _bSealed;         // Is this cube sealed yet?
    public int           _nDataSize;       // Size of the data stored in cube (i.e. rows)

    /** Iterator position **/
    public int           _nIterPos;        // Current index of the iterator
    public int           _nIterStart;      // Starting point for iterator under restrictions
    public int           _nIterEnd;        // Ending point for iterator under restrictions

    /** The following parameters are for printing only **/
    public int           _nTableWidth;     // Table width to use for printing data cubes
    public ArrayList     _alPrintDimOrder; // Dimension indices and order to print them
    public PrintStream   _ps;              // PrintStream for table printing
    public int           _nAgFun;          // Aggregation function ID (see DMQL.Display for IDs)
    public List          _lParams;         // List of paramters for above aggregation function
    public DecimalFormat _df = new DecimalFormat("#.####");

    /** Constructor - default
     *  @param dimensions Must be a list with the dimensions in the order
     *                    that they appear.
     **/
    public DataCube(ArrayList dimensions, Dimension value) {

	_alDimensions = dimensions; // Set the dimensions array
	_dimVal       = value;
	_bSealed = false;           // Need to start off accepting data!
	_nDataSize = 0;             // Position for next element
	_data    = new DataElement[INITIAL_SIZE];

	_nIterPos = 0;              // Reset the iterator position

	// Initialize printing parameters
	_alPrintDimOrder = null;
	_ps = System.out;
	_nAgFun = DMQL.Display.NO_AG;
	_lParams = null;

	// Now, initialize the Comparator for this object from
	// the size of the dimensions array (i.e. assuming
	// all dimensions initially participate in the DataCube)
	ArrayList dim_sort_order = new ArrayList();
	for (int count = 0; count < _alDimensions.size(); count++) {
	    ((Dimension)_alDimensions.get(count))._nIndex = count;
	    dim_sort_order.add(new Integer(count));
	}
	_comp = new DCComp(dim_sort_order);
    }

    //////////////////////////////////////////////////////////////////////
    //                Data Cube Input/Configuration Methods
    //////////////////////////////////////////////////////////////////////

    /** Add a new entry to the data cube (don't sort)
     *  @param key Array of dim keys (same order as _alDimensions)
     *  @param val Value for this key
     **/
    public void addElement(Integer[] key, Integer val) {

	// Make sure data cube is not sealed
	if (_bSealed) {
	    System.out.println("Cannot add data to a sealed data cube!");
	    System.exit(1);
	}

	// How big is our current array?
	int array_size = _data.length;
	
	// Double array if required - amortized constant time!
	if (_nDataSize == array_size) {
	    DataElement[] temp = new DataElement[array_size << 1];
	    System.arraycopy(_data, 0, temp, 0, _nDataSize);
	    _data = temp;
	}

	// Now add new element
	_data[_nDataSize++] = new DataElement(key, val);
    }

    /** Once this is executed, the data is sorted but no more elements
     *  can be added due to the computational complexity of doing so
     **/
    public void sealDataCube() {
	
	// Set the sealed flag
	_bSealed = true;

	// Shrink the array to just the right size
	DataElement[] temp = new DataElement[_nDataSize];
	System.arraycopy(_data, 0, temp, 0, _nDataSize);
	_data = temp;	
	
	// Now, sort the data according to the current comparator
	Arrays.sort(_data, _comp);
    }

    /** Resort the data according to the current Comparator (_comp)
     **/
    public void resortData() {

	// Check to see if sealed
	if (!_bSealed) {
	    System.out.println("Cannot sort an unsealed data cube!");
	    System.exit(1);
	}

	// Now, sort according to this comparator
	Arrays.sort(_data, _comp);

	// Find start and end indices according to restrictions
	_nIterStart = findRestrictionRange(false);
	_nIterEnd   = findRestrictionRange(true);
    }

    /** Adds a dimension to the DataCube and resort the DataCube
     *  @param dim_name The name of the dimension to add
     *  @return Whether successful
     **/
    public boolean drillDownDim(String dim_name) {

	// Add this key index to the Comparator (see DCComp methods)
	int index = getDimIndex(dim_name);
	if (index >= _alDimensions.size()) {
	    System.out.println("Dimension '" + dim_name + "' not found.");
	    return false;
	}
	_comp.addDim(new Integer(index));

	// Resort the DataCube!
	resortData();

	// Return whether successful
	return true;
    }

    /** Drops a dimension from the DataCube and resort the DataCube
     *  @param dim_name The name of the dimension to add
     *  @return Whether successful
     **/
    public boolean rollUpDim(String dim_name) {

	// Drop this key index from the Comparator (see DCComp methods)
	int index = getDimIndex(dim_name);
	if (index >= _alDimensions.size()) {
	    System.out.println("Dimension '" + dim_name + "' not found.");
	    return false;
	}
	_comp.dropDim(new Integer(index));

	// Resort the DataCube!
	resortData();

	// Return whether successful
	return true;
    }

    /** Adds a restriction to the DataCube and resort the DataCube
     *  @param dim_name The name of the dimension to add
     *  @return Whether successful
     **/
    public boolean addRestriction(String dim_name, Integer value) {

	// Add this key index to the Comparator (see DCComp methods)
	int index = getDimIndex(dim_name);
	if (index >= _alDimensions.size()) {
	    System.out.println("Dimension '" + dim_name + "' not found.");
	    return false;
	}
	_comp.addRestriction(new Integer(index), value);

	// Resort the DataCube!
	resortData();

	// Return whether successful
	return true;
    }

    /** Drops a restriction from the DataCube and resort the DataCube
     *  @param dim_name The name of the dimension to add
     *  @return Whether successful
     **/
    public boolean dropRestriction(String dim_name) {

	// Drop this key index from the Comparator (see DCComp methods)
	int index = getDimIndex(dim_name);
	if (index >= _alDimensions.size()) {
	    System.out.println("Dimension '" + dim_name + "' not found.");
	    return false;
	}
	_comp.dropRestriction(new Integer(index));

	// Resort the DataCube!
	resortData();

	// Return whether successful
	return true;
    }

    /** Finds dimension index
     *  @param dim_name The name of the dimension
     *  @return index of dimension, -1 if not found
     **/
    public int getDimIndex(String dim_name) {

	// Find the key index for the given dim_name
	// E.g., if _alDimensions=[DimAge, DimLoc] then DimAge has index 0,
	//       DimLoc has index 1.  Note that _alDimensions contains
	//       Dimension objects, not Strings!
	int index = 0;
	for (index = 0; index < _alDimensions.size(); index++) {
	    Dimension d = (Dimension)_alDimensions.get(index);
	    if (d._sDimName.equals(dim_name)) {
		return index;
	    }
	}

	return -1;
    }

    //////////////////////////////////////////////////////////////////////
    //                       Data Retrieval Methods
    //////////////////////////////////////////////////////////////////////

    /** Get a set of all dimensions (active, inactive, restricted)
     *  @return Set of Strings (dim names)
     **/
    public Set getAllDims() {
	HashSet restr = new HashSet();
	Iterator i = _alDimensions.iterator();
	while (i.hasNext()) {
	    Dimension d = (Dimension)i.next();
	    restr.add(d._sDimName);
	}

	return restr;
    }

    /** Get a set of the current active datacube restrictions
     *  @return Set of Strings (dim names)
     **/
    public Map getRestrictions() {
	HashMap restr = new HashMap();
	Iterator i = _comp._hmRestrictions.entrySet().iterator();
	while (i.hasNext()) {
	    Map.Entry me = (Map.Entry)i.next();
	    int index = ((Integer)me.getKey()).intValue();
	    restr.put(((Dimension)_alDimensions.get(index))._sDimName, me.getValue());
	}

	return restr;
    }

    /** Get a set of the current active datacube dimensions
     *  @return Set of Strings (dim names)
     **/
    public Set getActiveDims() {
	HashSet restr = new HashSet();
	Iterator i = _comp._alCompOrder.iterator();
	while (i.hasNext()) {
	    int index = ((Integer)i.next()).intValue();
	    restr.add(((Dimension)_alDimensions.get(index))._sDimName);
	}

	return restr;
    }

    /** Get the possible set of keys for a dimension
	@param dim_name Name of the dimension
     *  @return Set of keys (Integers)
     **/
    public Set getDimKeys(String dim_name) {
	int index = getDimIndex(dim_name);
	if (index < 0) {
	    return null;
	} else {
	    return ((Dimension)_alDimensions.get(index)).getDimKeys();
	}
    }

    /** Get a list of indices for the non-dropped Dimensions
     *  (in the order they are compared)
     *  @return List of indices (Integers) for non-dropped Dimensions
     **/
    public List getIndexList() {
	return (List)_comp._alCompOrder.clone();
    }

    /** Get the Dimension object for a key index
     *  @param index The key index for this DataCube
     *  @return The Dimension object for the given key index
     **/
    public Dimension getDim(int index) {
	return (Dimension)_alDimensions.get(index);
    }

    /** Get the size of this datacube 
     *  @return Number of elements in datacube
     **/
    public int size() {
	return _nDataSize;
    }

    /** Get all data under current restrictions
	@return List of data elements (could be empty)
     **/
    public List getRestrictedData() {
	//System.out.println(_nIterStart + "/" + _nIterEnd);
	ArrayList vals = new ArrayList();
   	resetIterator();
	while (hasNext()) {
	    DataSet ds = (DataSet)next();
	    vals.addAll( ds._values );
	}
	return vals;
    }

    /** Retrieves all values for the given key according to current
     *  comparator (Comp)
     *  @param key Array of key values in same order as _alDimensions
     *  @return List of values for the given key (empty list if nothing)
     **/
    public List getValues(Integer[] key) {

	// Find an index where this key occurs
	DataElement de_key = new DataElement(key, null);
	int index = Arrays.binarySearch(_data, de_key, _comp);
	ArrayList ret_list = new ArrayList();

	if (index >= 0) {

	    // Could be multiple entries... search backward for all values
	    int first = index;
	    while (first > 0) {
		if (_comp.compare(_data[first - 1], de_key) == 0) {
		    first--;
		} else {
		    break;
		}
	    }
	    
	    // Could be multiple entries... search forward for all values
	    int last = index;
	    while (last < (_nDataSize - 1)) {
		if (_comp.compare(_data[last + 1], de_key) == 0) {
		    last++;
		} else {
		    break;
		}
	    }
	    
	    // Now, build the set of matching values 
	    for (int i = first; i <= last; i++) {
		ret_list.add( ((DataElement)_data[i])._val );
	    }
	}
	
	// Return the value set
	return ret_list;
    }

    /** Find min or max key for data with current restrictions
     *  @param max if true will return max key, otherwise min key
     *  @return index of min or max key associated with current restrictions
     *
     *  This is inefficient... should do with binary search.
     **/
    public int findRestrictionRange(boolean max) {

	// Now advance iterator according to current restrictions
	Integer[] key = new Integer[_alDimensions.size()];
	for (int i = 0; i < _alDimensions.size(); i++) {
	    Integer val = null;
	    if ( (val = (Integer)_comp._hmRestrictions.get(new Integer(i))) != null) {
		key[i] = val;
	    } else {
		key[i] = max ? new Integer(Integer.MAX_VALUE) : new Integer(Integer.MIN_VALUE);
	    }
	}

	DataElement de_key = new DataElement(key, null);
	int index = max ? _nDataSize : 0;
	if (max) {
	    while (index > 0) {
		if (_comp.compare(de_key, _data[index-1]) < 0) {
		    index--;
		} else {
		    break;
		}
	    }
	} else {
	    while (index < _nDataSize) {
		if (_comp.compare(de_key, _data[index]) > 0) {
		    index++;
		} else {
		    break;
		}
	    }
	}

	if (_comp._hmRestrictions.size() == 0 && 
	    ((max == false && index != 0) ||
	     (max == true  && index != _nDataSize))) {
	    System.out.println("Error: [min,max] key should be full range");
	    System.exit(1);
	}

	return index;
    }

    /** Resets the iterator
     **/
    public void resetIterator() {
	_nIterPos = _nIterStart;
    }

    /** Returns whether there are more elements to iterate over
     *  @return True if more elements to iterate over
     **/
    public boolean hasNext() {
	return (_nIterPos < _nIterEnd);
    }

    /** Returns the next DataSet object for entries sharing the same key,
     *  starting at _nIterPos.  Note that the key should not be changed,
     *  it is not a copy but rather an actual key!  Assumes the array
     *  has been sorted for the current Comparator (_comp) used to test
     *  equality.
     *
     *  @return Next DataSet object (i.e. key and values)
     **/
    public DataSet next() {
	
	// Data left?
	if (_nIterPos >= _nIterEnd) {
	    return null;
	}

	// Retrieve the key
	DataSet ds = new DataSet(((DataElement)_data[_nIterPos])._key);
	DataElement de = new DataElement(ds._key, null);

	// Now retrieve the data as long as this key matches the data
	while (_nIterPos < _nIterEnd &&
	       _comp.compare(de, ((DataElement)_data[_nIterPos])) == 0) {
	    ds.addValue(((DataElement)_data[_nIterPos])._val);
	    _nIterPos++; // Move on to next element...
	}

	// DataSet built so return it
	return ds;
    }

    //////////////////////////////////////////////////////////////////////
    //                  Internal classes - data storage 
    //////////////////////////////////////////////////////////////////////

    /**
     *  Internal class for data storage 
     *  (not static because we want to access parent class fields)
     **/
    public class DataElement {

	/** Data Members **/
	public Integer[] _key;
	public Integer   _val;

	/** Constructor
	 *  @param key Key for this entry in the DataCube
	 *  @param val Value for this entry in the DataCube
	 **/
	public DataElement(Integer[] key, Integer val) {
	    _key = key;
	    _val = val;
	}

	/** Prints contents of object
	 *  @return String representation of object
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer("Key: [");
	    boolean first = true;
	    for (int n=0; n < _key.length; n++) {
		if (first) {
		    first = false;
		} else {
		    sb.append(" ");
		}
		sb.append(((Dimension)_alDimensions.get(n)).getStringVal(_key[n]));
	    }
	    sb.append("] -> Val: " + _val.toString());
	    return sb.toString();
	}
    }

    /**
     *  Internal class for iterator element storage
     *  (not static because we want to access parent class fields)
     **/
    public class DataSet {

	/** Data Members **/
	public Integer[] _key;
	public List      _values;

	/** Constructor
	 *  @param key Key for this entry in the DataCube
	 *  @param val Value for this entry in the DataCube
	 **/
	public DataSet(Integer[] key) {
	    _key    = key;
	    _values = new ArrayList();
	}

	/** Add an element to value set
	 *  @param val The element to add
	 **/
	public void addValue(Integer val) {
	    _values.add(val);
	}

	/** Prints contents of object
	 *  @return String representation of object
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer("Key: [");
	    boolean first = true;
	    for (int n=0; n < _key.length; n++) {
		if (first) {
		    first = false;
		} else {
		    sb.append(" ");
		}
		sb.append(((Dimension)_alDimensions.get(n)).getStringVal(_key[n]));
	    }
	    sb.append("] -> Val: " + _values.toString());
	    return sb.toString();
	}

    }

    //////////////////////////////////////////////////////////////////////
    //                     Printing and helper methods 
    //////////////////////////////////////////////////////////////////////

    /** Print this data cube to the given output stream
     *  @param ps Stream to print to
     *  @param table_width Max width of a table element
     *  @param ag_fun int ID of aggregation function to use (see DMQL.Display)
     *  @param params List of parameters for the aggregation function
     **/
    public void printCube(PrintStream ps, int table_width, int ag_fun, List params) {

	// Set the table width and stream for printing
	_ps = ps;
	_nTableWidth = table_width;
	_nAgFun = ag_fun;
	_lParams = params;

	// For now, just print the dimensions in the order given
	// by the Comparator (which of course lacks dropped dims)
	_alPrintDimOrder = _comp._alCompOrder;

	// Make up a key to be reused and call method to print,
	// starting at the 0th element of _alPrintDimOrder
	Integer[] key = new Integer[_alDimensions.size()];
	printDimension(0, key);
    }

    /** Recurse on the dimensions and print them
     *  @param dims List of Dimensions (in order)
     *  @param index Dimension to print
     **/
    public void printDimension(int index, Integer[] key) {

	if (index >= _alPrintDimOrder.size()) {

	    // No indices, just one value (at least, assumed min of one value exists!)
	    key = _data[0]._key;
	    String cell_data = computeAgVal(getValues(key), _nAgFun, _lParams);
	    _ps.println("\nSingle cell: " + cell_data);

	} else if (index == (_alPrintDimOrder.size() - 1)) {

	    // Print out a list for single dimension
	    int dim_index = ((Integer)_alPrintDimOrder.get(index)).intValue();
	    Dimension d = (Dimension)_alDimensions.get(dim_index);
	    Iterator dim_vals = d.getIndices().iterator();
	    while (dim_vals.hasNext()) {
		Integer keyval = (Integer)dim_vals.next();
		key[dim_index] = keyval;
		String cell_data = computeAgVal(getValues(key), _nAgFun, _lParams);
		_ps.print("\n" + indent(index) + "* " + d._sDimName + "=" + 
				 d.getStringVal(keyval) + ": " + cell_data);
	    }
	    _ps.println();

	} else if (index == (_alPrintDimOrder.size() - 2)) {
	    
	    // Print out the table for last two dims
	    int dim_index1 = ((Integer)_alPrintDimOrder.get(index)).intValue();
	    int dim_index2 = ((Integer)_alPrintDimOrder.get(index+1)).intValue();
	    Dimension d1 = (Dimension)_alDimensions.get(dim_index1);
	    Dimension d2 = (Dimension)_alDimensions.get(dim_index2);
	    Iterator id1 = d1.getIndices().iterator();
	    Iterator id2 = d2.getIndices().iterator();
	    Integer d1_val, d2_val;

	    // Print headers
	    _ps.print("\n" + indent(index) + trunc(d1._sDimName, (_nTableWidth-1)/2) + 
			     "\\" + trunc(d2._sDimName, (_nTableWidth-1)/2 + ((_nTableWidth-1) % 2))  + " |");
	    StringBuffer line = new StringBuffer(repeat("-",_nTableWidth+1) + "+-");
	    while (id2.hasNext()) {
		d2_val = (Integer)id2.next();
		_ps.print(" " + trunc(d2.getStringVal(d2_val), _nTableWidth));
		line.append(repeat("-",_nTableWidth+1));
	    }
	    _ps.print("\n" + indent(index) + line);

	    // Print contents
	    while (id1.hasNext()) {
		d1_val = (Integer)id1.next();
		key[dim_index1] = d1_val;
		_ps.print("\n" + indent(index) + trunc(d1.getStringVal(d1_val), _nTableWidth) + " |");
		id2 = d2.getIndices().iterator();
		while (id2.hasNext()) {
		    d2_val = (Integer)id2.next();
		    key[dim_index2] = d2_val; 
		    String cell_data = trunc(computeAgVal(getValues(key), _nAgFun, _lParams), _nTableWidth);
		    _ps.print(" " + cell_data);
		}
	    }
	    _ps.println();

	} else {
	    
	    // Iterate over the current dim and recurse on the next
	    int dim_index = ((Integer)_alPrintDimOrder.get(index)).intValue();
	    Dimension d = (Dimension)_alDimensions.get(dim_index);
	    Iterator dim_vals = d.getIndices().iterator();
	    while (dim_vals.hasNext()) {
		Integer keyval = (Integer)dim_vals.next();
		key[dim_index] = keyval;
		_ps.println("\n" + indent(index) + "* " + d._sDimName + "=" + d.getStringVal(keyval) + ": ");
		printDimension(index+1, key);
	    }
	}

    }

    /** Given a data list (of Integers), an aggregation function, a
     *  list of fun params, and width, returns a String resulting from
     *  the application of the specified function to the list
     *
     *  @param data List of data
     *  @param ag_fun int ID of the aggregation function (see DMQL.Display)
     *  @param params List of params for the aggregation function
     *
     *  @return String representation of function result
     **/
    public String computeAgVal(List data, int ag_fun, List params) {

	if (data.size() == 0) {
	    return "~"; // No possible function computation!
	}

	switch (ag_fun) {
	case DMQL.Display.NO_AG: {
	    return data.toString();

	} 
	case DMQL.Display.SUM: {
	    int sum = 0;
	    Iterator i = data.iterator();
	    while (i.hasNext()) {
		sum += ((Integer)i.next()).intValue();
	    }
	    return "[" + Integer.toString(sum) + "]";
		
	} 
	case DMQL.Display.MIN: {
	    Iterator i = data.iterator();
	    int min = ((Integer)i.next()).intValue();
	    while (i.hasNext()) {
		int val = ((Integer)i.next()).intValue();
		if (val < min) {
		    min = val;
		}
	    }
	    return "[" + Integer.toString(min) + "]";

	} 
	case DMQL.Display.MAX: {
	    Iterator i = data.iterator();
	    int max = ((Integer)i.next()).intValue();
	    while (i.hasNext()) {
		int val = ((Integer)i.next()).intValue();
		if (val > max) {
		    max = val;
		}
	    }
	    return "[" + Integer.toString(max) + "]";

	} 
	case DMQL.Display.AVG: {
	    int sum = 0;
	    Iterator i = data.iterator();
	    while (i.hasNext()) {
		sum += ((Integer)i.next()).intValue();
	    }
	    //System.out.println("" + sum + " " + data.size() + " " + sum/data.size());
	    return "[" + _df.format(((double)sum)/((double)data.size())) + "]";

	} 
	case DMQL.Display.COUNT_ALL: {
	    return "[" + Integer.toString(data.size()) + "]";

	} 
	case DMQL.Display.COUNT_VAL: {
	    int count = 0;
	    String param = (String)params.get(0);
	    int pval = 0;
	    try {
		pval = Integer.parseInt(param);
	    } catch (NumberFormatException e) {
		return "?" + param;
	    }
	    Iterator i = data.iterator();
	    while (i.hasNext()) {
		if ( ((Integer)i.next()).intValue() == pval) {
		    count++;
		}
	    }
	    if (count > 0) {
		return "[" + count + "]";
	    } else {
		return "~";
	    }

	} 

	default: {
	    return "?" + ag_fun;
	}
	}
    }

    //////////////////////////////////////////////////////////////////////////////
    //                         Miscellaneous Helper Methods
    //////////////////////////////////////////////////////////////////////////////
    
    /** A simple method for truncating (or padding) string to a length
     *  @param s   String to process
     *  @param len Length to pad/truncate to
     *  @return    Truncated/Padded version of string
     **/
    public static String trunc(String s, int len) {
	if (s.length() <= len) {
	    return pad(s, len);
	} else {
	    return s.substring(0, len);
	}
    }

    /** A simple method for producing an indent based on level
     *  @param level Level to indent for
     *  @return String of spaces for indentation
     **/
    public static String indent(int level) {
	StringBuffer sb = new StringBuffer();
	for (int i=0; i<level; i++) {
	    sb.append("   ");
	}
	return sb.toString();
    }

    /** Produces repeated String
     *  @param s String to repeat
     *  @param n number of times to repeat
     *  @return Repeated String
     **/
    public static String repeat(String s, int n) {
	StringBuffer sb = new StringBuffer();
	for (int i=0; i<n; i++) {
	    sb.append(s);
	}
	return sb.toString();
    }

    /** A simple method for padding Strings to a given length 
     *  @param s   String to pad
     *  @param len Length to pad to
     *  @return    Padded version of string
     **/
    public static String pad(String s, int len) {
	for (int i=s.length(); i<len; i++) {
	    s += " ";
	}
	return s;
    }

}
