//////////////////////////////////////////////////////////////////////
//
// File:     Domain.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     10/1/2003
//
//////////////////////////////////////////////////////////////////////

// Package definition
package dm.datacube;

// Packages to import
import dm.db.*;
import dm.dmql.*;
import dm.mine.*;
import java.sql.*;
import java.util.*;

/**
 * Implements a Domain using an array and a DCComp object
 *
 * @version   1.0
 * @author    Scott Sanner
 * @language  Java (JDK 1.3)
 **/
public class Domain {

    /** Local data members **/
    protected String      _sDomName;            // The name of this domain
    protected String      _sDriver;             // The driver String for the database
    protected String      _sDatabase;           // The name of the database to use
    protected DBInterface _db;                  // The DBInterface for this domain!
    protected HashMap     _hmDimName2DefineDim; // A mapping from dim names to definitions
    protected HashMap     _hmCubeName2DataCube; // A mapping from cube names to DataCubes

    /** Constructor
     *  @param dom_name Name to call this domain (can be anything)
     *  @param driver   Driver string for SQL
     *  @param db_name  DB name for SQL
     **/
    public Domain(String dom_name, String driver, String db_name) {

	// Initialize local vars
	_sDomName  = dom_name;
	_sDriver   = driver;
	_sDatabase = db_name;

	// Initialize data structs
	_hmDimName2DefineDim = new HashMap();
	_hmCubeName2DataCube = new HashMap();

	// Open the DB Connection
	_db = new DBInterface();
	_db.open(_sDriver, _sDatabase);
    }

    /** Process a query by dispatching to correct method
     *  @param dc The DMQL command to process
     *  @return Whether the command was successfully processed
     **/
    public boolean process(DMQLCommand dc) {
	if (dc == null) {
	    return false;
	} else if (dc instanceof DMQL.DefineDim) {
	    return processDefineDim((DMQL.DefineDim)dc);

	} else if (dc instanceof DMQL.DefineCube) {
	    return processDefineCube((DMQL.DefineCube)dc);

	} else if (dc instanceof DMQL.AddDropDim) {
	    return processAddDropDim((DMQL.AddDropDim)dc);

	} else if (dc instanceof DMQL.Display) {
	    return processDisplay((DMQL.Display)dc);	    

	} else if (dc instanceof DMQL.FindClassRules) {
	    return processFindClassRule((DMQL.FindClassRules)dc);

	} else {
	    System.out.println("\nCommand class: '" + dc.getClass() + 
			       "' not implemented");
	    return false;
	}
    }

    /** Finalize - make sure to close database connection on finalize
     **/
    protected void finalize() {

	// Close the DB Connection
	_db.close();
    }

    //////////////////////////////////////////////////////////////////////////////
    //                DMQL Command Implementation: "define dimension"
    //////////////////////////////////////////////////////////////////////////////

    /** Process a specific DMQL command
     *  @param com The DMQLCommand subclass to process
     *  @return Whether the processing was successful
     **/
    public boolean processDefineDim(DMQL.DefineDim com) {

	System.out.println("\nProcessing Dimension Query:\n");
	System.out.println(com);

	// This is simple... just cache the dimension definition
	// according to it's name... we'll use it later.
	_hmDimName2DefineDim.put(com._sDimName, com);

	return true;
    }

    //////////////////////////////////////////////////////////////////////////////
    //                  DMQL Command Implementation: "define cube"
    //////////////////////////////////////////////////////////////////////////////

    /** Process a specific DMQL command
     *  @param com The DMQLCommand subclass to process
     *  @return Whether the processing was successful
     **/
    public boolean processDefineCube(DMQL.DefineCube com) {

	System.out.println("\nProcessing Cube Query:\n");
	System.out.println(com);
	DataCube dc = null;

	// Note to students: This already contains an implementation
	// for some "dummy data".  Your job is to process the actual
	// command by querying the backend database and populating the
	// Dimension and DataCube objects with the retrieved data.
	// See the helper methods (below) that are called from this method.

	// Now, go through each of the dimensions in the query and
	ArrayList dimensions = new ArrayList();
	if (processDimensions(com, dimensions)) {

	    // Initialize a new DataCube
	    Dimension value = (Dimension)dimensions.remove(dimensions.size()-1);
	    dc = new DataCube(dimensions, value);

	} else {

	    // There was an error
	    return false;
	}
	
	// Now, go through the cube data and process it
	if (processCubeData(dc, com)) {

	    // Map the name of this cube to the actual cube so that other
	    // DMQL commands can retrieve it
	    _hmCubeName2DataCube.put(com._sCubeName, dc);

	} else {

	    // There was an error
	    return false;
	}

	// Done and successful
	return true;
    }
    
    /** Process and build the dimension data for the DataCube - helper method for 
     *  processDefineCube
     *  @param  dc The datacube
     *  @param  com The DMQL.DefineCube command
     *  @return Whether the processing was successful
     **/
    public boolean processDimensions(DMQL.DefineCube com, ArrayList dims) {
	
	// Get an iterator for the dimensions
	ArrayList augmented_dims = new ArrayList();
	augmented_dims.addAll(com._alDims);
	augmented_dims.add(com._sValDim);
	Iterator dim_iter = augmented_dims.iterator();

	// Now, populate each dimension and add it to dims... note that
	// this maintains the key ordering given in the command.
	while (dim_iter.hasNext()) {

	    // Get the Dimension definition for the dim name, make a
	    // new Dimension, and add it to the dims list
	    String dim_name = (String)dim_iter.next();
	    DMQL.DefineDim ddim = (DMQL.DefineDim)_hmDimName2DefineDim.get(dim_name);
	    Dimension d = new Dimension(dim_name);
	    dims.add(d);

	    // Now, populate the Dimension from ddim and com by
	    // querying the database (_db)
	    
	    // Construct the query
	    StringBuffer cols = new StringBuffer(ddim._sKey);
	    int num_cols = ddim._alCols.size();
	    Iterator   i = ddim._alCols.iterator();
	    while (i.hasNext()) {
		cols.append(", " + i.next());
	    }

	    // Query the database for the dim data
	    ResultSet rs = _db.query("SELECT " + cols.toString() + 
				     " FROM "  + ddim._sTable);
	    if (rs == null) {
		return false;
	    }
	    
	    // Add the dim data to the Dimension object
	    try {
		while (rs.next()) { 

		    // Process the non-key column data for each row
		    ArrayList dimdata = new ArrayList();
		    for (int col = 2; col <= (num_cols+1); col++){ 
			dimdata.add(rs.getString(col));
		    }
		    
		    // Add the dimension data for this row
		    d.addDimData(new Integer(Integer.parseInt(rs.getString(1))), 
				 dimdata);
		}
	    } catch (SQLException e) {
		System.out.println("Error: " + e);
		return false;
	    } catch (NumberFormatException e) {
		System.out.println("Error: " + e);
		return false;
	    }
	}	

	return true;
    }

    /** Process the actual DataCube data - helper method for processDefineCube
     *  @param  dc The datacube
     *  @param  com The DMQL.DefineCube command
     *  @return Whether successful
     **/
    public boolean processCubeData(DataCube dc, DMQL.DefineCube com) {

	// Construct the query
	StringBuffer cols = new StringBuffer();
	int num_cols = com._alKeys.size();
	for (int col = 0; col < num_cols; col++) {
		cols.append(com._alKeys.get(col) + ", ");
	}
	
	// Query the database for the cube data
	ResultSet rs = _db.query("SELECT " + cols.toString() + com._sValKey + 
				 " FROM "  + com._sTable);
	if (rs == null) {
	    return false;
	}

	// Add each row to the data cube
	try {
	    while (rs.next()) {

		// Get the key
		Integer[] key = new Integer[num_cols];
		for (int col = 0; col < num_cols; col++){
		    key[col] = new Integer(Integer.parseInt(rs.getString(col + 1)));
		}

		// Add the key and value data
		dc.addElement(key, new Integer(
                              Integer.parseInt(rs.getString(num_cols + 1))));
	    }
	} catch (SQLException e) {
	    System.out.println("Error: " + e);
	    return false;
	} catch (NumberFormatException e) {
	    System.out.println("Error: " + e);
	    return false;
	}

	// Last step in populating DataCube is to seal it so that it 
	// becomes sorted.
	dc.sealDataCube();

	return true;
    }

    //////////////////////////////////////////////////////////////////////////////
    //              DMQL Command Implementation: "add/drop dimension"
    //////////////////////////////////////////////////////////////////////////////
 
    /** Process a specific DMQL command
     *  @param com The DMQLCommand subclass to process
     *  @return Whether the processing was successful
     **/
    public boolean processAddDropDim(DMQL.AddDropDim com) {

	System.out.println("\nProcessing Display Query:\n");
	System.out.println(com);

	// Retrieve the data cube
	DataCube dc = (DataCube)_hmCubeName2DataCube.get(com._sCubeName);
	if (dc == null) {
	    System.out.println("\nData cube not found...");
	    return false;
	}
	
	// Perform the add/drop function and return result
	if (com._bAdd) { //AddDim if true
	    return dc.drillDownDim(com._sDimName);
	} else {         //DropDim if false
	    return dc.rollUpDim(com._sDimName);
	}
    }

    //////////////////////////////////////////////////////////////////////////////
    //                 DMQL Command Implementation: "display cube"
    //////////////////////////////////////////////////////////////////////////////

    /** Process a specific DMQL command
     *  @param com The DMQLCommand subclass to process
     *  @return Whether the processing was successful
     **/
    public boolean processDisplay(DMQL.Display com) {

	System.out.println("\nProcessing Display Query:\n");
	System.out.println(com);

	// Retrieve the data cube
	DataCube dc = (DataCube)_hmCubeName2DataCube.get(com._sCubeName);
	if (dc == null) {
	    System.out.println("\nData cube not found...");
	    return false;
	}

	// Now, print the data cube
	dc.printCube(System.out, com._nTableWidth, com._nAgFunction, com._alAgParams);

	return true;
    }

    //////////////////////////////////////////////////////////////////////////////
    //           DMQL Command Implementation: "find classification rules"
    //                 Need to implement new mining algorithm class
    //////////////////////////////////////////////////////////////////////////////

    /** Process a specific DMQL command
     *  @param com The DMQLCommand subclass to process
     *  @return Whether the processing was successful
     **/
    public boolean processFindClassRule(DMQL.FindClassRules com) {

	System.out.println("\nProcessing FindClassRule Query:\n");
	System.out.println(com);

	// Retrieve the data cube
	DataCube dc = (DataCube)_hmCubeName2DataCube.get(com._sCubeName);
	if (dc == null) {
	    System.out.println("\nData cube not found...");
	    return false;
	}

	// Now, mine the classification rules using the parameters and algorithm
	// class defined in this com instance
	List rules = new OLAPRuleLearn().mineClassRules(com, dc);

	// Display the rules!
	Iterator i = rules.iterator(); 
	System.out.println("\nThe following rules met the query requirements: ");
	if (!i.hasNext()) {
	    System.out.print("\n  * [No rules found]");
	}
	while (i.hasNext()) {
	    System.out.print("\n" + i.next());
	}
	System.out.println();

	return true;
    }

}
