//////////////////////////////////////////////////////////////////////
//
// File:     DBInterface.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     10/1/2003
//
// Description:
//
//   A simple database interface.  All database interaction can be
//   handled with the following methods:
//
//   - void open(String driver, String dbname)
//   - boolean ResultSet query(String sql_query)
//   - void close()
//
//   Additionally, the contents of a ResultSet can be printed with
//   the following static method:
//
//   - static void PrintResultSet(ResultSet rs)
//
//////////////////////////////////////////////////////////////////////

// Package definition
package dm.db;

// Packages to import
import java.sql.*;

/** 
 * Our simple database interface class
 **/
public class DBInterface {

    /** Database specific objects - you won't need to use these directly **/
    public Statement  _s;
    public Connection _con;

    /** Constructor 
     **/
    public DBInterface() {

	try {
	    // Load the JDBC driver
	    Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");

	} catch (Exception e) {
	    System.out.println("Error loading driver: " + e);
	}

    }

    /** Opens a database 
     *  @param driver Name of JDBC/ODBC driver
     *  @param dbname Name of database to access via driver
     **/
    public void open(String driver, String dbname) {

	try {
	    
	    // Get a connection to the MS Access database specified by dbname
	    _con = DriverManager.getConnection(driver + dbname, "", "");
	    System.out.println("\n- Database \"" + dbname + "\" successfully opened.");

	    // Create a Statement object for performing queries
	    _s = _con.createStatement();

	} catch (Exception e) {
	    System.out.println("Error: " + e);
	}

    }

    /** Performs a query 
     *  @param  query A String containing a SQL query
     *  @return A ResultSet for the query, essentially a table.  See
     *          DBInterface.PrintResultSet(...) for how to view and 
     *          access this data.
     **/
    public ResultSet query(String query) {

	ResultSet rs = null;

	try {

	    // Execute a query on the database and obtain a result set
	    _s.execute(query);
	    System.out.println("\n- Executing query: \"" + query + "\"");

	    rs = _s.getResultSet();

	} catch (Exception e) {
	    System.out.println("Error: " + e);
	}

	return rs;
    }

    /** Close the database **/
    public void close() {
	try {

	    // Close all database connections
	    _s.close();
	    _con.close();
	    System.out.println("\n- Database successfully closed.");

	} catch (Exception e) {
	    System.out.println("Error: " + e);
	}

    }

    /** Print a ResultSet (static) 
     *  @param rs ResultSet to print to System.out
     **/
    public static void PrintResultSet(ResultSet rs) {

	if (rs == null) {
	    System.out.println("\nInvalid ResultSet");
	    return;
	}

	try {

	    // Get and print column headers
	    ResultSetMetaData md = rs.getMetaData();
	    int count = md.getColumnCount();
	    System.out.println();
	    for (int c=1; c<=count; c++) {
		System.out.print(pad(md.getColumnName(c), 9) + " ");
	    }
	    
	    System.out.println("\n================================" + 
			       "==================================");
	    
	    // Print all column data for each row
	    while (rs.next()) {
		for (int c=1; c<=count; c++) {
		    System.out.print(pad(rs.getString(c), 9) + " ");
		}
		System.out.println();
	    }

	} catch (Exception e) {
	    System.out.println("Error: " + e);
	}

    }

    /** A simple method for padding Strings to a given length 
     *  @param s   String to pad
     *  @param len Length to pad to
     *  @return    Padded version of string
     **/
    public static String pad(String s, int len) {
	for (int i=s.length(); i<len; i++) {
	    s += " ";
	}
	return s;
    }
    
    /** A simple exmaple using a Microsoft Access database connection and 
     *  a general SQL query **/
    public static void main(String args[]) {

	// Make a new database object
	DBInterface db = new DBInterface();

	// Connect to the specified database
	db.open("jdbc:odbc:Driver={Microsoft Access Driver (*.mdb)};DBQ=",
		"h:/JDBC/StarTest1.mdb");

	// Get the result of an example query
	ResultSet rs = db.query("SELECT * FROM MainData;");

	// Print this result set (it's just a table).
	// Note: we can only print the ResultSet once...
	// if we want to print it again, we'll have to 
	// perform the query a second time.
	PrintResultSet(rs);

	// Close the database connection
	db.close();
    }

}
    
