//////////////////////////////////////////////////////////////////////
//
// File:     DMQL.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     10/1/2003
//
//////////////////////////////////////////////////////////////////////

// Package definition
package dm.dmql;

// Packages to import
import dm.mine.*;
import java.io.*;
import java.lang.*;
import java.util.*;

/**
 * Just a wrapper class for DMQL
 *
 * @version   1.0
 * @author    Scott Sanner
 * @language  Java (JDK 1.3)
 **/
public class DMQL
{
    /** Process the initial portion of a DMQL String in an attempt
     *  to pass it off to a DMQLCommand subclass for processing
     *  @param  query The DMQL query String
     *  @return A DMQLCommand instance or null (if error during parse)
     **/
    public static DMQLCommand ProcessQuery(String query) {

	DMQLCommand dcom = null;

	// Set up the TokenStreams... one to read, one to pass
	TokenStream ts = new RawTextTokenStream(); 
	TokenStream ts2 = new RawTextTokenStream();

	try {
	    ts.process(query);
	    ts2.process(query);
	    
	    // Determine which DMQL command to dispatch this query to for
	    // further processing...
	    Token t = ts.nextToken();
	    if (t != null) {
		if (t._sToken.equalsIgnoreCase("define")) {
		    t = ts.nextToken();
		    if (t._sToken.equalsIgnoreCase("dimension")) {
			dcom = new DefineDim();
			if (!dcom.parse(ts2)) dcom = null;
		    } else if (t._sToken.equalsIgnoreCase("cube")) {
			dcom = new DefineCube();
			if (!dcom.parse(ts2)) dcom = null;
		    }
		} else if (t._sToken.equalsIgnoreCase("add") ||
			   t._sToken.equalsIgnoreCase("drop")) {
		    dcom = new AddDropDim();
		    if (!dcom.parse(ts2)) dcom = null;
		} else if (t._sToken.equalsIgnoreCase("display")) {
		    dcom = new Display();
		    if (!dcom.parse(ts2)) dcom = null;
		} else if (t._sToken.equalsIgnoreCase("find")) {
		    dcom = new FindClassRules();
		    if (!dcom.parse(ts2)) dcom = null;
		}
	    }
	} catch (Exception e) {
	    System.out.println("Error while processing query: " + e);
	    dcom = null;
	} 
	
	return dcom;
    }

    //////////////////////////////////////////////////////////////////
    //                       DMQLCommand Subtypes
    //////////////////////////////////////////////////////////////////

    /** 
     * A DMQLCommand query subtype 
     **/
    public static class DefineDim extends DMQLCommand {

	/** Local members for this DMQL command - contents should be 
	 ** obvious from command syntax **/
	public String    _sDimName;
	public String    _sTable;
	public String    _sKey;
	public ArrayList _alCols;
	
	/** Default constructor
	 **/
	public DefineDim() {
	    _alCols = new ArrayList();
	}

	/** Parse this DMQL command from the given TokenStream
	 *  @param  ts The TokenStream containing the DMQL command
	 *  @return True is the parse was successful, false otherwise
	 **/
	public boolean parse(TokenStream ts) {

	    // System.out.println("\nProcessing DMQL for " + this.getClass());

	    try {
		ts.match("define");
		ts.match("dimension");
		_sDimName = ts.nextToken()._sToken;
		ts.match("from");
		_sTable = ts.nextToken()._sToken;
		ts.match("as");
		_sKey = ts.nextToken()._sToken;
		while (ts.nextTokenEquals(",")) {
		    ts.match(",");
		    _alCols.add(ts.nextToken()._sToken);
		}

	    } catch (TokenStreamException e) {
		System.out.println("\nParse Error: " + e);
		return false;
	    }

	    return true;
	}

	/** Default constructor
	 *  @return String representation of this command
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append(  "* DMQL.DefineDim");
	    sb.append("\n  - Dim:   " + _sDimName);
	    sb.append("\n  - Table: " + _sTable);
	    sb.append("\n  - Key:   " + _sKey);
	    sb.append("\n  - Cols:  " + _alCols);
	    return sb.toString();
	}
    }

    /** 
     * A DMQLCommand query subtype 
     **/
    public static class DefineCube extends DMQLCommand {

	/** Local members for this DMQL command - contents should be 
	 ** obvious from command syntax **/
	public String    _sCubeName;
	public String    _sTable;
	public ArrayList _alKeys;
	public ArrayList _alDims;
	public String    _sValKey;
	public String    _sValDim;

	/** Default constructor
	 **/
	public DefineCube() {
	    _alKeys = new ArrayList();
	    _alDims = new ArrayList();
	}

	/** Parse this DMQL command from the given TokenStream
	 *  @param  ts The TokenStream containing the DMQL command
	 *  @return True is the parse was successful, false otherwise
	 **/
	public boolean parse(TokenStream ts) {

	    // System.out.println("\nProcessing DMQL for " + this.getClass());

	    try {
		ts.match("define");
		ts.match("cube");
		_sCubeName = ts.nextToken()._sToken;
		ts.match("from");
		_sTable = ts.nextToken()._sToken;
		ts.match("with");
		ts.match("dimensions");
		_alKeys.add(ts.nextToken()._sToken);
		ts.match("for");
		_alDims.add(ts.nextToken()._sToken);
		while (ts.nextTokenEquals(",")) {
		    ts.match(",");
		    _alKeys.add(ts.nextToken()._sToken);
		    ts.match("for");
		    _alDims.add(ts.nextToken()._sToken);
		}
		ts.match("having");
		ts.match("content");
		_sValKey = ts.nextToken()._sToken;
		ts.match("for");
		_sValDim = ts.nextToken()._sToken;

	    } catch (TokenStreamException e) {
		System.out.println("\nParse Error: " + e);
		return false;
	    }

	    return true;
	}

	/** Default constructor
	 *  @return String representation of this command
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append(  "* DMQL.DefineCube");
	    sb.append("\n  - Name:   " + _sCubeName);
	    sb.append("\n  - Table:  " + _sTable);
	    sb.append("\n  - Keys:   " + _alKeys);
	    sb.append("\n  - Dims:   " + _alDims);
	    sb.append("\n  - ValKey: " + _sValKey);
	    sb.append("\n  - ValDim: " + _sValDim);
	    return sb.toString();
	}
    }

    /** 
     * A DMQLCommand query subtype 
     **/
    public static class AddDropDim extends DMQLCommand {

	/** Local members for this DMQL command - contents should be 
	 ** obvious from command syntax **/
	public String    _sCubeName;
	public String    _sDimName;
	public boolean   _bAdd; // If not add then drop

	/** Default constructor
	 **/
	public AddDropDim() {
	    _bAdd = true;
	}

	/** Parse this DMQL command from the given TokenStream
	 *  @param  ts The TokenStream containing the DMQL command
	 *  @return True is the parse was successful, false otherwise
	 **/
	public boolean parse(TokenStream ts) {

	    // System.out.println("\nProcessing DMQL for " + this.getClass());

	    try {
		Token t = ts.nextToken();
		_bAdd = t._sToken.equalsIgnoreCase("add");
		_sDimName = ts.nextToken()._sToken;
		if (_bAdd) {
		    ts.match("to");
		} else {
		    ts.match("from");
		}
		_sCubeName = ts.nextToken()._sToken;

	    } catch (TokenStreamException e) {
		System.out.println("\nParse Error: " + e);
		return false;
	    }

	    return true;
	}

	/** Default constructor
	 *  @return String representation of this command
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append(  "* DMQL.AddDropDim");
	    sb.append("\n  - Fun:  " + (_bAdd ? "add" : "drop"));
	    sb.append("\n  - Dim:  " + _sDimName);
	    sb.append("\n  - Cube: " + _sCubeName);
	    return sb.toString();
	}
    }

    /** 
     * A DMQLCommand query subtype 
     **/
    public static class Display extends DMQLCommand {

	/** Local members for this DMQL command - contents should be 
	 ** obvious from command syntax **/
	public String    _sCubeName;
	public String    _sAgFunction;
	public int       _nAgFunction;
	public ArrayList _alAgParams;
	public int       _nTableWidth;

	/** Constants for aggregation functions **/
	public final static int NO_AG     = 0;
	public final static int SUM       = 1;
	public final static int MIN       = 2;
	public final static int MAX       = 3;
	public final static int AVG       = 4;
	public final static int COUNT_ALL = 5;
	public final static int COUNT_VAL = 6;

	/** Default constructor
	 **/
	public Display() {
	    _alAgParams = new ArrayList();
	    _nTableWidth = 10;
	}

	/** Parse this DMQL command from the given TokenStream
	 *  @param  ts The TokenStream containing the DMQL command
	 *  @return True is the parse was successful, false otherwise
	 **/
	public boolean parse(TokenStream ts) {

	    // System.out.println("\nProcessing DMQL for " + this.getClass());

	    try {
		ts.match("display");
		_sCubeName = ts.nextToken()._sToken;

		if (ts.nextTokenEquals("with")) {
		    ts.match("with");
		    ts.match("width");
		    String num = ts.nextToken()._sToken;
		    try {
			_nTableWidth = Integer.parseInt(num);
		    } catch (NumberFormatException e) {
			throw new TokenStreamException("Invalid width:" + num);
		    }
		}
		
		if (ts.nextTokenEquals("using")) {

		    // The user-specified aggregation method
		    ts.match("using");
		    _sAgFunction = ts.nextToken()._sToken;

		    // Collect all parameters
		    Token t;
		    while ((t = ts.nextToken()) != null) {
			_alAgParams.add(t._sToken);
		    }

		    // Now, match the function
		    if (_sAgFunction.equalsIgnoreCase("sum")) {
			_nAgFunction = SUM;
		    } else if (_sAgFunction.equalsIgnoreCase("min")) {
			_nAgFunction = MIN;
		    } else if (_sAgFunction.equalsIgnoreCase("max")) {
			_nAgFunction = MAX;
		    } else if (_sAgFunction.equalsIgnoreCase("avg")) {
			_nAgFunction = AVG;
		    } else if (_sAgFunction.equalsIgnoreCase("count-all")) {
			_nAgFunction = COUNT_ALL;
		    } else if (_sAgFunction.equalsIgnoreCase("count-val")) {
			_nAgFunction = COUNT_VAL;
			if (_alAgParams.size() != 1) {
			    throw new TokenStreamException("Required exactly 1 parameter for count-val.");
			}
		    } else {
			throw new TokenStreamException("Aggregation function '" + _sAgFunction + "' unknown.");
		    }

		} else {

		    // Make sure no more tokens
		    Token t = ts.nextToken();
		    if (t != null) {
			throw new TokenStreamException("Expected 'using' or end of string: " + t);
		    }
		    _sAgFunction = "none";
		    _nAgFunction = NO_AG;
		    
		}

	    } catch (TokenStreamException e) {
		System.out.println("\nParse Error: " + e);
		return false;
	    }

	    return true;
	}

	/** Default constructor
	 *  @return String representation of this command
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append(  "* DMQL.Display");
	    sb.append("\n  - Cube:   " + _sCubeName);
	    sb.append("\n  - Ag-fun: " + _sAgFunction + " (" + _nAgFunction + ")");
	    sb.append("\n  - Params: " + _alAgParams);
	    return sb.toString();
	}
    }

    /** 
     * A DMQLCommand query subtype 
     **/
    public static class FindClassRules extends DMQLCommand {

	/** Local members for this DMQL command - contents should be 
	 ** obvious from command syntax **/
	public String    _sCubeName;
	public ArrayList _alDims;
	public double    _dSupport;
	public double    _dConfidence;
	public Class     _cClassAlg;
	public ClassificationProc _cp;

	/** Default constructor
	 **/
	public FindClassRules() {
	    _alDims = new ArrayList();
	    _cClassAlg = null;
	    _cp = null;
	}

	/** Parse this DMQL command from the given TokenStream
	 *  @param  ts The TokenStream containing the DMQL command
	 *  @return True is the parse was successful, false otherwise
	 **/
	public boolean parse(TokenStream ts) {

	    // System.out.println("\nProcessing DMQL for " + this.getClass());

	    try {
		ts.match("find");
		ts.match("classification");
		ts.match("rules");
		ts.match("for");
		_sCubeName = ts.nextToken()._sToken;
		ts.match("related");
		ts.match("to");
		_alDims.add(ts.nextToken()._sToken);
		while (ts.nextTokenEquals(",")) {
		    ts.match(",");
		    _alDims.add(ts.nextToken()._sToken);
		}
		ts.match("with");
		ts.match("support");
		ts.match("threshold");
		_dSupport = Double.parseDouble(ts.nextToken()._sToken);
		ts.match(",");
		ts.match("confidence");
		ts.match("threshold");
		_dConfidence = Double.parseDouble(ts.nextToken()._sToken);

	    } catch (TokenStreamException e) {
		System.out.println("\nParse Error: " + e);
		return false;
	    }

	    return true;
	}

	/** Sets the algorithm class for this command
	 *  @param String name of algorithm class
	 **/
	public boolean setClassAlg(String class_name) {
	    try {
		_cClassAlg = Class.forName(class_name);
		_cp = (ClassificationProc)_cClassAlg.newInstance();
	    } catch (Exception e) {
		return false;
	    }
	    
	    return true;
	}

	/** Default constructor
	 *  @return String representation of this command
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append(  "* DMQL.FindClassRules");
	    sb.append("\n  - Cube:     " + _sCubeName);
	    sb.append("\n  - Dims:     " + _alDims);
	    sb.append("\n  - Support:  " + _dSupport);
	    sb.append("\n  - Confid:   " + _dConfidence);
	    return sb.toString();
	}
    }

}
