/////////////////////////////////////////////////////////////////////////
//
// File:     RawTextTokenStream.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     9/1/2003
//
// Notes: 
//
// 1) Supports primitive top-down parsing with match(String str) and
// next token lookahead nextTokenEquals(String str).
//
// 2) Currently only supports ',' as a special symbol token, all
// others treated as part of token or whitespace.  This can be changed
// without too much difficulty... just need to add code for caching
// any special symbols...  see how comma is handled and extend from
// there.
//
/////////////////////////////////////////////////////////////////////////

// Package definition
package dm.dmql;

// Packages to import
import java.io.*;
import java.util.*;

/**
 * Tokenizes a generic text document
 *
 * @version   1.0
 * @author    Scott Sanner
 * @language  Java (JDK 1.3)
 **/
public class RawTextTokenStream
    implements TokenStream {

    /** Static Constants **/
    protected static final int EOF = -1;

    /** Non-static Data Members **/
    protected BufferedReader       _brInput;
    protected int                  _nLine;
    protected int                  _nLinePos;
    protected int                  _nPos;
    protected boolean              _bHoldCommaForNext;
    protected Token                _tLookAhead;

    /** Constructor - default
     **/
    public RawTextTokenStream() {
	_brInput      = null;
	_nPos         = -1; // No chars read yet
	_nLinePos     = -1; // No chars read yet
	_nLine        = 1;  // Start at line 1
    }

    /** Process a String for tokenizing
     *  @param str String for tokenizing
     **/
    public void process(String str) 
	throws TokenStreamException {

	// Initialize components
	_nPos = _nLinePos = -1; // No chars read yet
	_nLine = 1;             // Start at line 1
	_bHoldCommaForNext = false;
	_tLookAhead = null;
	if (_brInput != null) {
	    close();
	}

	_brInput = new BufferedReader(new StringReader(str));
    }

    /** Close the file currently being tokenized
     **/
    public void close() 
	throws TokenStreamException {
	
	// Close the reader and reset 
	try {
	    if (_brInput != null) {
		_brInput.close();
		_brInput = null;
	    }

	} catch (IOException ex) {
	    throw new TokenStreamException("IOException: " + ex, 
					   _nPos, _nLine, _nLinePos);
	}
    }

    /** Attempts to match the next token, throws TokenStreamException 
     *  @param content String to match next token against
     **/
    public void match(String content) throws TokenStreamException {
	Token t = nextToken();
	if (t == null || !t._sToken.equalsIgnoreCase(content)) {
	    throw new TokenStreamException("Expected: '" + content + "' from " + t);
	}
    }


    /** Determines if the next token is the specified String
     *  @param  str  The String to be compared against
     *  @return true if next token is equal to str
     **/
    public boolean nextTokenEquals(String str) throws TokenStreamException {
	if (_tLookAhead == null) {
	    _tLookAhead = nextTokenInternal();
	}
	if (_tLookAhead == null) {
	    return false;
	}
	return _tLookAhead._sToken.equalsIgnoreCase(str);
    }
    
    /** Returns the next Token from a TokenStream or null
     *  if the end of the stream has been reached.  Throws
     *  a TokenStreamException if there is an IO error.
     *  @return The next token or null (if no tokens remain)
     **/    
    public Token nextToken() 
	throws TokenStreamException {

	if (_tLookAhead != null) {
	    Token ret_val =_tLookAhead;
	    _tLookAhead = null;
	    return ret_val;
	} else {
	    return nextTokenInternal();
	}
    }
    
    /** Returns the next Token from a TokenStream or null
     *  if the end of the stream has been reached.  Throws
     *  a TokenStreamException if there is an IO error.
     *  @return The next token or null (if no tokens remain)
     **/    
    public Token nextTokenInternal() 
	throws TokenStreamException {

	StringBuffer token = new StringBuffer();
	char[] cbuf = new char[1];
	int token_len, cur_line_pos, cur_line;

	// Read a character at a time from the current line,
	// return on EOF or when current token complete
	while (true) {

	    if (_bHoldCommaForNext) {
		_bHoldCommaForNext = false;
		return new Token(",", _nPos, _nLine, _nLinePos);
	    }

	    // Read a character, check for EOF
	    try {
		if (_brInput.read(cbuf, 0, 1) == EOF) {
		    token_len = token.length();
		    if (token_len > 0) {
			return new Token(token.toString(),
					 _nPos - token_len,
					 _nLine, _nLinePos - token_len); 		    
		    } else {
			return null;
		    }
		}
		++_nPos; 

		// Use temp line pos to retain current position
		// in case a newline resets it, otherwise position
		// will be calculated incorrectly when token returned
		cur_line = _nLine;
		cur_line_pos = ++_nLinePos;
		
	    } catch (IOException ex) {
		throw new TokenStreamException("IOException: " + ex, 
					       _nPos, _nLine, _nLinePos);
	    }

	    // Handle newline
	    if (cbuf[0] == '\n') {
		++_nLine;
		_nLinePos = -1;
	    }

	    //////////////////////////////////////////////////////////
	    //    Handle token chars, whitespace, special symbols
	    //////////////////////////////////////////////////////////

	    // This is a valid, non-tag character so decide what to
	    // do...  namely, if char is a delimiter then return the
	    // current token if it exists.
	    if (!Character.isLetterOrDigit(cbuf[0]) && 
		!("-.*%\"'&()~;".indexOf(cbuf[0]) >= 0)) {
		
		// A delimiter so return token if non-empty,
		// otherwise ignore character and continue.
		// Note special handling for ',' - it is a
		// token by itself.
		token_len = token.length();
		if (token_len > 0) {
		    
		    if (cbuf[0] == ',') {
			_bHoldCommaForNext = true;
		    }

		    return new Token(token.toString(),
				     _nPos - token_len,
				     cur_line, cur_line_pos - token_len);

		} else if (cbuf[0] == ',') {
		    return new Token(",", _nPos, _nLine, _nLinePos); 
		}

	    } else {

		// Character is a valid part of token
		token.append(cbuf[0]);
	    }	    

	    //////////////////////////////////////////////////////////
	}
    }

}
