////////////////////////////////////////////////////////////////////////
//
// File:     OLAPRuleLearn.java
// Project:  MIE457F Data Mining System
// Author:   Scott Sanner, University of Toronto (ssanner@cs.toronto.edu)
// Date:     10/1/2003
//
// Description:
//
//   Implements a Decision Tree Classification Procedure using the
//   Info Gain metric.  (Inefficient, see notes below for improvements.)
//
////////////////////////////////////////////////////////////////////////

// Package definition
package dm.mine;

// Packages to import
import dm.dmql.*;
import dm.datacube.*;
import java.text.*;
import java.util.*;

/**
 * Implements a Decision Tree Classification Procedure using OLAP
 * operations on a DataCube.
 *
 * @version   1.0
 * @author    Scott Sanner
 * @language  Java (JDK 1.3)
 **/
public class OLAPRuleLearn 
    implements ClassificationProc {

    /** Local data members **/
    public double    _dMinSupport;      // Min support threshold for a rule
    public double    _dMinConf;         // Min confidence threshold for a rule
    public List      _alLegalSplitDims; // Legal dimensions for a rule antecedent
    public DataCube  _dc;               // The data cube!

    /** Perform a data mining procedure to find classification rules
     *  using decision tree learning and the Gini metric.
     *  Uses the parameters of com and operates on DataCube d.
     *
     *  @param  com The command defining the paramters of this 
     *              classification procedure
     *  @param  d   The datacube to mine
     *  @return A List object containing rules.  The type of this
     *          object is undefined.  At the very least, it must simply
     *          implement the toString() method so that its rule
     *          contents can be displayed for the user.
     **/
    public List mineClassRules(DMQL.FindClassRules com, DataCube d) {

	// Initialize local vars
	_dc          = d;
	_dMinSupport = com._dSupport;
	_dMinConf    = com._dConfidence;
	_alLegalSplitDims = com._alDims;

	// Roll-up the datacube into a 0-D cuboid
	Iterator i = _dc.getAllDims().iterator();
	while (i.hasNext()) {
	    String dim_name = (String)i.next();
	    _dc.rollUpDim(dim_name);
	    _dc.dropRestriction(dim_name);
	}

	// Perform the data mining
	List rules = mineClassRules();

	// Return data cube to full n-D cuboid (i.e., drill down)
	i = _dc.getAllDims().iterator();
	while (i.hasNext()) {
	    String dim_name = (String)i.next();
	    _dc.drillDownDim(dim_name);
	}

	// Return the list of mined rules
	return rules;
    }

    /** Perform a data mining procedure to find classification rules
     *  using decision tree learning based on OLAP operations.
     *
     *  This procedure searches all of the remaining dimensions,
     *  i.e. those elements in SplitDims which have not been
     *  restricted, and determines which one yields the highest
     *  information gain.  The procedure then restricts on each of
     *  these split values and recursively calls itself.
     *
     *  Before returning, it determines if a rule should be added at
     *  this level of the decision tree and adds it to the return
     *  list.
     *
     *  @return A List object containing Rules (see the internal class
     *          at the end of this file)
     **/    
    protected List mineClassRules() {

	// Initialize rules set
	ArrayList rules = new ArrayList();

	// Check to see if we are out of dimensions to split on
	Set current_splits = _dc.getRestrictions().keySet();
	if (current_splits.size() >= _alLegalSplitDims.size()) {
	    return rules; // No more splits, return empty rules list
	}

	// Obtain the pre-split data set and determine the size
	List   pre_split_data      = _dc.getRestrictedData();
	int    pre_split_data_size = pre_split_data.size();
	if (pre_split_data_size == 0) {
	    return rules; // No data to mine!
	}

	// Initialize best split info gain and dim
	double best_info_gain = -1;
	String best_split_dim = null;

	// Iterate through all remaining legal dimensions for splitting
	// and keep track of the one with the highest info gain.  
	Iterator split_dim_iterator = _alLegalSplitDims.iterator();
	while (split_dim_iterator.hasNext()) {
	    String split_dim = (String)split_dim_iterator.next();

	    // Skip this dimension if we've split on it already
	    if (current_splits.contains(split_dim)) {
		continue;
	    }
	    
	    // Initialize info gain to pre_split entropy
	    double info_gain = calcEntropy(pre_split_data);
	    
	    // Split on each dimension value and update info gain
	    Set dim_keys = _dc.getDimKeys(split_dim);
	    Iterator split_val_iterator = dim_keys.iterator();
	    while (split_val_iterator.hasNext()) {
		
		// Restrict split dimension to the split value
		Integer split_val = (Integer)split_val_iterator.next();
		_dc.addRestriction(split_dim, split_val);
		
		// Get data subset for current split value (if any)
		List   post_split_data      = _dc.getRestrictedData();
		int    post_split_data_size = post_split_data.size();
		if (post_split_data_size != 0) {
		
		    // Calculate info_gain decrement for this split
		    double post_split_entropy   = calcEntropy(post_split_data);
		    info_gain -= ((double)post_split_data_size/pre_split_data_size)*post_split_entropy;
		}
		    
		// Need to undo restriction before continuing
		_dc.dropRestriction(split_dim);
	    }

	    // Is this the current best split?
	    if (info_gain > best_info_gain) {
		best_info_gain = info_gain;
		best_split_dim = split_dim;
	    }
	}

	// Continue only if some split yielded a positive gain
	if (best_info_gain <= 0.0d) {
	    return rules; 
	}

	// Now we know the best split, restrict data for each split value
	// and recurse on the sub-splits of the decision tree.
	Set dim_keys = _dc.getDimKeys(best_split_dim);
	Iterator split_val_iterator = dim_keys.iterator();
	while (split_val_iterator.hasNext()) {
	    
	    // Restrict split dimension to the split value
	    Integer split_val = (Integer)split_val_iterator.next();
	    _dc.addRestriction(best_split_dim, split_val);
	    
	    // Recurse on restricted data and add returned rules to current list
	    rules.addAll(mineClassRules());
	    
	    // Generate a rule for this split (if it is not empty)
	    List split_data = _dc.getRestrictedData();
	    int total_split_data = split_data.size();
	    if (total_split_data != 0) {

		// Get the subset of data for this split and it's value frequencies
		HashMap val2freq = calcValueFrequencies( split_data );

		// Iterate over each possible rule consequent value and compile rule
		Iterator freq_iter = val2freq.entrySet().iterator();
		while (freq_iter.hasNext()) {
		
		    Map.Entry me   = (Map.Entry)freq_iter.next();
		    Integer   val  = (Integer)me.getKey();
		    int       freq = ((Integer)me.getValue()).intValue();

		    // Confidence is the (#tuples matching both condition & value class)
		    //                   -----------------------------------------------
		    //                             (#tuples matching condition)
		    double confidence = (double)freq / (double)total_split_data;

		    // Support is the              (#tuples matching condition)
		    //                   ----------------------------------------------
		    //                                 (#tuples in datacube) 
		    double support    = (double)total_split_data / (double)_dc.size();

		    // Add the rule only if we meet the criteria
		    if (confidence >= _dMinConf && support >= _dMinSupport) {
			rules.add( new Rule(support /* rule support - not being used currently */, 
					    confidence            /* rule confidence */,
					    _dc.getRestrictions() /* conditions */,
					    val                   /* consequent */));
		    }
		}
	    }

	    // Need to undo restriction before continuing
	    _dc.dropRestriction(best_split_dim);
	}

	// We're done with this node of the decision tree, so return rules learned
	return rules;
    }

    /** Determines the entropy metric for a list of values.
     *
     *  This method simply calculates the frequency of each value in the
     *  list and uses this information to calculate the entropy of the
     *  data.
     *
     *  @param c Collection of data
     *  @return The value of the Info metric for this data collection
     **/
    protected double calcEntropy(Collection c) {

	// Determine total number of values in list
	int    total = c.size();
	double info  = 0.0d;
	
	// Get the frequency count of each value in list
	HashMap val2freq = calcValueFrequencies(c);
	
	// Now that we've computed frequecies, let's compute the Info metric
	Iterator freq_iter = val2freq.entrySet().iterator();
	while (freq_iter.hasNext()) {

	    Map.Entry me   = (Map.Entry)freq_iter.next();
	    int       freq = ((Integer)me.getValue()).intValue();
	    
	    double p_i  = (double)freq / (double)total;
	    info       -= p_i * Math.log( p_i );
	}

	// We're done... return the value!
	return info;
    }

    /** Determines the gini metric for a list of values.
     *
     *  This method simply calculates the frequency of each value in the
     *  list and uses this information to calculate the gini metric for the
     *  data.
     *
     *  @param c Collection of data
     *  @return The value of the gini metric for this data collection
     **/
    protected double calcGini(Collection c) {

	// Determine total number of values in list
	int    total = c.size();
	double gini  = 0.0d;
	
	// Get the frequency count of each value in list
	HashMap val2freq = calcValueFrequencies(c);
	
	// Now that we've computed frequecies, let's compute the Info metric
	Iterator freq_iter = val2freq.entrySet().iterator();
	while (freq_iter.hasNext()) {

	    Map.Entry me   = (Map.Entry)freq_iter.next();
	    int       freq = ((Integer)me.getValue()).intValue();
	    
	    double p_i  = (double)freq / (double)total;
	    gini       -= p_i * p_i;
	}

	// We're done... return the value!
	return gini;
    }

    /** Determines the entropy metric for a list of values.
     *
     *  This method simply calculates the frequency of each value in the
     *  list and returns an array of frequencies for each value.
     *
     *  @param l Collection of data
     *  @return HashMap Maps from list-values (Integer) to their frequency (Integer)
     **/
    protected HashMap calcValueFrequencies(Collection c) {
	
	// For each element, increment the correct value count
	HashMap val2freq = new HashMap();
	Iterator i = c.iterator();
	while (i.hasNext()) {
	    Integer cur_val  = (Integer)i.next();
	    Integer cur_freq = (Integer)val2freq.get(cur_val);
	    if (cur_freq == null) {
		val2freq.put(cur_val, new Integer(1));
	    } else {
		val2freq.put(cur_val, new Integer(cur_freq.intValue()+1));		
	    }
	}
	
	// Return the value -> freq map
	return val2freq;
    }

    /////////////////////////////////////////////////////////////////////////////
    //                            Internal Rule Class
    /////////////////////////////////////////////////////////////////////////////

    /** 
     *  Internal class to represents a data mining rule 
     **/
    public class Rule {

	/** Static data members **/
	public DecimalFormat _df = new DecimalFormat("#.####");

	/** Data members**/
	public double        _dSupport;
	public double        _dConf;
	public ArrayList     _alDims;
	public ArrayList     _alDimValues;
	public String        _sConseq;

	/** Rule constructor
	 *  @param support      Support value for this rule
	 *  @param confidence   Confidence value for this rule
	 *  @param ant_dims     Antecedent dimension indices
	 *  @param ant_vals     Antecedent value selections (match dims)
	 *  @param conseq_label Consequent label (database column)
	 *  @param consequent   Consequent value
	 **/
	public Rule(double support, double confidence, 
		    Map condition_dim_vals,
		    Integer consequent) {

	    _dSupport     = support;
	    _dConf        = confidence;
	    _alDims       = new ArrayList();
	    _alDimValues  = new ArrayList();
	    _sConseq      = _dc._dimVal.getStringVal(consequent); 

	    Iterator i = condition_dim_vals.entrySet().iterator();
	    while (i.hasNext()) {
		Map.Entry me     = (Map.Entry)i.next();
		String  dim_name = (String)me.getKey();
		Integer value    = (Integer)me.getValue();

		_alDims.add(_dc._alDimensions.get(_dc.getDimIndex(dim_name)));
		_alDimValues.add(value);
	    }
	}

	/** Convert this rule to String format
	 *  @return String version of this rule
	 **/
	public String toString() {
	    StringBuffer sb = new StringBuffer();
	    sb.append("  * Rule: ");
	    Iterator id = _alDims.iterator();
	    Iterator iv = _alDimValues.iterator();
	    boolean first = true;
	    while (id.hasNext()) {
		Dimension d = (Dimension)id.next();
		Integer   v = (Integer)iv.next();
		if (first) {
		    first = false;
		} else {
		    sb.append(" ^ ");
		}
		sb.append(/*d._sDimName + "=" +*/ d.getStringVal(v));
	    }
	    if (_alDims.size() == 0) {
		sb.append("TRUE");
	    }
	    sb.append(" => " + _sConseq);
	    if (_dSupport != -1d) {
		sb.append("  [Sup: " + _df.format(_dSupport) + "]");
	    }
	    if (_dConf != -1d) {	
		sb.append("  [Conf: " + _df.format(_dConf) + "]");
	    }
	    return sb.toString();
	}
    }    

}
